@@ -1,5 +1,6 @@
 from typing import List
 
+import torch
 import torch.nn as nn
 from torch import Tensor
 from transformers import Wav2Vec2Processor, Wav2Vec2Model
@@ -14,6 +15,9 @@ class UpstreamExpert(nn.Module):
         self.processor = Wav2Vec2Processor.from_pretrained(ckpt)
         self.model = Wav2Vec2Model.from_pretrained(ckpt)
 
+    def get_downsample_rates(self, key: str) -> int:
+        return 320
+
     def forward(self, wavs: List[Tensor]):
         device = wavs[0].device
         processor_outputs = self.processor(
@@ -22,9 +26,12 @@ class UpstreamExpert(nn.Module):
             sampling_rate=SAMPLE_RATE,
             padding="longest",
         )
+        attention_mask = processor_outputs.get("attention_mask", None)
+        if isinstance(attention_mask, torch.Tensor):
+            attention_mask = attention_mask.to(device)
         model_outputs = self.model(
             processor_outputs.input_values.to(device),
-            attention_mask=processor_outputs.get("attention_mask", None),
+            attention_mask=attention_mask,
             output_hidden_states=True,
         )
         return {
